from datetime import date, timedelta

from fastapi import APIRouter, Depends, Security
from sqlalchemy import and_, or_, select
from sqlalchemy.orm import Session, joinedload

from ...core.security import get_current_active_user as get_user
from ...db.session import SessionFuture
from ...models.menu_category import MenuCategory
from ...models.product_version import ProductVersion
from ...schemas.user_token import UserToken
from . import report_finish_date, report_start_date


router = APIRouter()


@router.get("")
def product_updates_report_view(
    start_date: date = Depends(report_start_date),
    finish_date: date = Depends(report_finish_date),
    user: UserToken = Security(get_user, scopes=["product-sale-report"]),
):
    with SessionFuture() as db:
        return {
            "startDate": start_date.strftime("%d-%b-%Y"),
            "finishDate": finish_date.strftime("%d-%b-%Y"),
            "report": product_updates_report(start_date, finish_date, db),
        }


def product_updates_report(s: date, f: date, db: Session) -> list[str]:
    list_ = (
        db.execute(
            select(ProductVersion)
            .join(ProductVersion.menu_category)
            .where(
                or_(
                    and_(
                        ProductVersion.valid_from >= s,
                        ProductVersion.valid_from <= f,
                    ),
                    and_(
                        ProductVersion.valid_till >= s - timedelta(days=1),
                        ProductVersion.valid_till <= f,
                    ),
                )
            )
            .order_by(
                MenuCategory.sort_order,
                MenuCategory.name,
                ProductVersion.sort_order,
                ProductVersion.name,
                ProductVersion.valid_from.nullsfirst(),
            )
            .options(
                joinedload(ProductVersion.menu_category, innerjoin=True),
            )
        )
        .scalars()
        .all()
    )
    report = {}
    for item in list_:
        if item.product_id not in report:
            report[item.product_id] = ""
        report[item.product_id] += (
            "From: "
            + (f"{item.valid_from:%d-%b-%Y}" if item.valid_from is not None else "\u221e")
            + " "
            + "Till: "
            + (f"{item.valid_till:%d-%b-%Y}" if item.valid_till is not None else "\u221e")
            + "\n"
            + f"{item.full_name} @ {item.price: .2f} - {'Happy Hour' if item.has_happy_hour else 'No Happy Hour'}\n"
        )
    return list(report.values())
